/**
 * \file pathTool.hpp
 *
 * Header file for the rednav path tool.
 *
 * \date 2011-01-07
 * \author redouane.boumghar@laas.fr
 * \ingroup rednav
 */


#ifndef REDNAV_LIBS_PATH_TOOL_HPP
#define REDNAV_LIBS_PATH_TOOL_HPP

// For the definition of jafar::lgl::NavPath
#include <lgl/NavGraph.hpp>
#include <lgl/Attributes.hpp>


namespace jafar {
namespace rednav {

	typedef enum { 
		PATH_TWO_POINTS=1,
		PATH_THREE_POINTS,
		PATH_FOUR_POINTS,
		PATH_COMMON_FILTER,   // Apply subsampling and max number of points
		PATH_RDV_CROP,
		PATH_TWO_POINTS_LIMIT,
		PATH_DISTANCE_LIMIT,
		PATH_INTERPOLATION,
		PATH_SAFESAMPLING,
		PATH_SMARTSAMPLING
	} PATH_TRANSFORMATION;

	class PathTool {
		private:
		//{{{
			/** Subsampling parameter in meters
			 *
			 * Only positive values. Zero means no subsampling. While using
			 * the subsampling filter paths are subsampled with regards to the
			 * distance 'subsampler' between points.
			 */
			double subsampler;

			/** Indicate how many way points the output path can contain
			 *
			 * Only positive values. Zero means no filter on points number.
			 * 1 means only the first point after the start point so two real
			 * points.
			 *
			 */
			int maxwaypts;

			/** Limit on binary entropy (free-obstacle knowledge) */
			double pathMaxEntropy;

			/** Limit on obstacle probability - path safety */
			double pathMaxProba;

			/** For calculation in meter we need this resolution factor */
			double cellScale;

			/** For location map-registration we need the zero-ref X coordinate */
			double geoXRoot;

			/** For location map-registration we need the zero-ref Y coordinate */
			double geoYRoot;

			/** Maximum distance (m) used for PATH_DISTANCE_LIMIT filter */
			double pathMaxDistance;

			/** Maximum distance (m) between allowed between waypoints  */
			double pathMaxStepDistance;

			/** Maximum deviation distance (m) allowed on a portion of a path  */
			double pathMaxDevDistance;

			/** Sensor radius we are working with */
			double sensorRadius;

			/** First master point index where to start path interpolation
			 * @see pathInterpolation()
			 */
			int interpolationStartPointIdx;

			/** Second master point index where to stop path interpolation
			 * @see pathInterpolation()
			 */
			int interpolationStopPointIdx;

			/** Interpolation distance */
			double interpolationDistance;

			/** NavGraph used to make some topological calculations */
			jafar::lgl::NavGraph* navgraph;
		//}}}

		public:
		//{{{
			/** Default constructor to use tools separatly */
			PathTool();

			/** One Shot constructor to apply a path transformation on a given path */
			PathTool(jafar::lgl::NavPath&, PATH_TRANSFORMATION, jafar::lgl::NavGraph* ng_=NULL);

			/** Destructor */
			~PathTool();

			/** Generic Tool to make easy generic calls 
			 *
			 * If you use this function, the filters will be called with the
			 * default values held by the intern parameters. Nevertheless they
			 * can be changed thanks to their respective setters.
			 *
			 * @param src Original path
			 * @param out Resulting path
			 * @param patid Path transformation identification
			 */
			bool applyTool(const jafar::lgl::NavPath& src, jafar::lgl::NavPath& out, PATH_TRANSFORMATION patid);

			/** Two points transformation */
			bool twoPoints(const jafar::lgl::NavPath& src, jafar::lgl::NavPath& out);

			/** Path subsampling */
			bool subSample(const jafar::lgl::NavPath& src, jafar::lgl::NavPath& out, double subsamp);

			/** Crop a path using multiple parameters
			 *
			 * @param vent Maximum admissible entropy value for YES-NO traversability
			 * @param vproba Maximum admissible obstacle probability
			 * @return true if a crop was done else false.
			 */
			bool cropPath(double vent, double vproba, jafar::lgl::NavPath& navpath);

			/** Crop a path using internal parameters and working on a path copy */
			bool cropPath(const jafar::lgl::NavPath& srcPath, jafar::lgl::NavPath& outPath);

			/** Two points transformation 
			 * The two points are chosen by comparing the distance between the
			 * straight track from the goal and the real track distance for
			 * each way point in the path.
			 * When at a certain waypoint the difference between these
			 * distances is superior to the sensor radius we choose this
			 * waypoint as the second point in the two points path.
			 */
			bool twoPointsFromCurve(const jafar::lgl::NavPath& src, jafar::lgl::NavPath& out);

			/** Two points transformation 
			 */
			bool cropDistance(const jafar::lgl::NavPath& src, jafar::lgl::NavPath& out, double cropDistance);

			/** Path interpolation 
			 *
			 * @param src original path
			 * @param out output path 
			 * @param smallestInterval wished interval between each interpolation (default interpolationDistance)
			 * @param masterP1 first index of the path from where interpolation is applied (default interpolationStartPointIdx)
			 * @param masterP2 second index of the path at where interpolation is stopped (default interpolationStopPointIdx)
			 *
			 */
			bool pathInterpolation(const jafar::lgl::NavPath& src, jafar::lgl::NavPath& out, double smallestInterval, int masterP1, int masterP2);

			/** Get only important points for trajectory
			 *
			 * Important points are defined as points were the trajectory has a
			 * great change in heading.
			 *
			 */
			bool pathSafeSampling(const jafar::lgl::NavPath& src, jafar::lgl::NavPath& out);

			/** Get important and mandatory points for trajectory
			 *
			 * It does as in pathSafeSampling and more by keeping points
			 * that are surounded by obstacles
			 */
			bool pathSmartSampling(const jafar::lgl::NavPath& src, jafar::lgl::NavPath& out);

			//-----------------------------------------------------/

			/** Transformation integer location to double map coordinates 
			 *
			 * Here we do not register the coordinates. So they can only be
			 * used for relative calculations like distances.
			 */
			double navLocToMeters(int a);


			/** Returns the meter distance between two navigation positions */
			double navLocDistance(const jafar::lgl::NavLocation& A, const jafar::lgl::NavLocation& B);

			/** Return the othogonal distance of C to segment AB. */
			double getOrthogonalHeight(const jafar::lgl::NavLocation& A, const jafar::lgl::NavLocation& B, const jafar::lgl::NavLocation& C);

			/** Return true if obstacles are close
			 *
			 * close means less than 3.0*pathMaxDevDistance
			 *
			 * @param First Obstacle Distance
			 */
			bool surroundingsDanger(const jafar::lgl::NavLocation& A, float& fodist);

			//-----------------------------------------------------/
			//--- set stuffs 

			/** Set the max number of way points */
			void setMaxWayPts(int n);

			/** Set the maximum binary (free-obst) entropy allowed for a path
			 *
			 * @see pathMaxEntropy
			 */
			void setPathMaxEntropy(double pme);

			/** Set the maximum obstacle probability allowed for a path
			 *
			 * @see pathMaxProba
			 */
			void setPathMaxProba(double pmp);

			/** Set the cell resolution
			 *
			 * @param cellres new cell resolution to set
			 */
			void setScale(double cellres);

			/** Set the map geo-X coordinate origin */
			void setGeoXRoot(double gxroot);

			/** Set the map geo-Y coordinate origin */
			void setGeoYRoot(double gyroot);

			/** Changes the subsampling factor */
			void setSubSampler(double setsb);

			/** Set the path maximum distance 
			 * @param pmd New distance in meters
			 * @see pathMaxDistance
			 */
			void setPathMaxDistance(double pmd);

			/** Set the path step maximum distance
			 * @param pmd New distance in meter
			 * @see pathMaxStepDistance
			 */
			void setPathMaxStepDistance(double pmd);

			/** Set the path maximum deviation distance
			 * @param pmd New distance in meter
			 * @see pathMaxDevDistance
			 */
			void setPathMaxDevDistance(double pmd);

			/** set the sensor radius parameter 
			 * @see sensorRadius
			 */
			void setSensorRadius(double ssd);
			
			/** Set the first point index for interpolation 
			 * @see interpolationStartPointIdx
			 */
			void setInterpolationStartPointIdx(int);

			/** Set the last point index for interpolation 
			 * @see interpolationStopPointIdx
			 */
			void setInterpolationStopPointIdx(int);

			/** Set the interpolation distance
			 * @see interpolationDistance
			 */
			void setInterpolationDistance(double);


			/** Tell which navigation graph to use */
			void setNavGraph(jafar::lgl::NavGraph* ng);

			//-----------------------------------------------------/
			//--- get stuffs 

			/** Return the pathMaxEntropy */
			double getPathMaxEntropy();

			/** Return the pathMaxProba */
			double getPathMaxProba();

			/** Return the subsampler */
			double getSubSampler();

			/** Get the path maximum distance in meter
			 * @see pathMaxDistance
			 */
			double getPathMaxDistance();

			/** Get the path maximum distance in meter
			 * @see pathMaxDistance
			 */
			double getPathMaxStepDistance();

			/** Get the path maximum deviation distance in meter
			 * @see pathMaxDevDistance
			 */
			double getPathMaxDevDistance();

			/** Get the sensor radius parameter 
			 * @see sensorRadius
			 */
			double getSensorRadius();

			/** Returns cell scale */
			double getScale();

			/** Return the geoX coordinate */
			double getGeoXRoot();

			/** Return the geoY coordinate */
			double getGeoYRoot();

			/** Get the last point index for interpolation */
			int getInterpolationStartPointIdx();

			/** Get the last point index for interpolation */
			int getInterpolationStopPointIdx();
		//}}}

	};
}} // end of namespaces

#endif //REDNAV_LIBS_PATH_TOOL_HPP
