/**
 * \file navigator.cpp
 *
 * Source file for the navigator tool.
 *
 * \date 2011-01-09
 * \author redouane.boumghar@laas.fr
 * \ingroup rednav
 */

#include <rednav/navigator.hpp>

namespace jafar {
namespace rednav {

	// Constructor that does nothing but giving calls to tools
	Navigator::Navigator():
		xsize(0),
		ysize(0),
		zsize(0)
	{ // {{{
		setGoal(0, 0, 0);
		setGoal(0.0, 0.0, 0.0);
		setStart(0, 0, 0);
		setStart(0.0, 0.0, 0.0);
	} // }}}

	// Static map file Constructor
	Navigator::Navigator(const std::string &inputfile)
	{ //{{{
		setGoal(0, 0, 0);
		setGoal(0.0, 0.0, 0.0);
		setStart(0, 0, 0);
		setStart(0.0, 0.0, 0.0);
		// Creating map
		navMap = new jafar::lgl::GeoData(inputfile);
		if (navMap != NULL)
		{
			// Setting up Geodata parameters
			setMapSize(navMap->getxsize(), navMap->getysize(), 0);
			pixelScale = navMap->getScale();
			// Creating the geo-decomposition
			navQuadtree = new jafar::lgl::Quadtree<double, jafar::lgl::GeoData>(navMap);
			// Creating the navigation graph
			navGraph = new jafar::lgl::NavGraph(navMap, navQuadtree);
			multiTool = new PathTool();
			multiTool->setScale(pixelScale);
			multiTool->setGeoXRoot(navMap->getxroot());
			multiTool->setGeoYRoot(navMap->getyroot());
			multiTool->setNavGraph(navGraph);
		}
	} //}}}

	// Tooled Constructor
	Navigator::Navigator(double* bufferedData, int xsize_, int ysize_, int zsize_, double xMapOrigin_, double yMapOrigin_, double zMapOrigin_, double pixelScale_)
	{ //{{{
		multiTool = new PathTool();
		multiTool->setScale(pixelScale_);

		setMapOrigin(xMapOrigin_, yMapOrigin_, zMapOrigin_, pixelScale_);
		setMapSize(xsize_, ysize_, zsize_);

		navMap = new jafar::lgl::GeoData(xsize, ysize, bufferedData);
		navQuadtree = new jafar::lgl::Quadtree<double, jafar::lgl::GeoData>(navMap);
		navGraph = new jafar::lgl::NavGraph(navMap, navQuadtree);
		multiTool->setNavGraph(navGraph);
	} //}}}

	//-------------------------- Destructor ---------
	Navigator::~Navigator()
	{ // {{{
		delete navGraph;
		delete navQuadtree;
		delete navMap;
	} // }}}

	//---- Functions

	bool Navigator::isUp()
	{//{{{
		if (navMap == NULL || navQuadtree == NULL || navGraph == NULL)
		{
			return false;
		}
		return true;
	}//}}}

	bool Navigator::savePath(const std::string& outfile,  jafar::lgl::NavPath& savingpath)
	{/*{{{*/
		// Saving the  path in a file
		typedef jafar::lgl::NavPath::const_iterator navpath_iterator;
		jafar::lgl::NavLocation curPos;
		std::ofstream pathStream (outfile.c_str());
		for( navpath_iterator lit = savingpath.begin(); lit != savingpath.end(); ++lit) {
			curPos =  ((*lit).first);
			pathStream << (curPos.utm_x + xMapOrigin) << "    " << (curPos.utm_y + yMapOrigin) << "    " 
					   << curPos.pos_x << "    " << curPos.pos_y << "    " 
			           << curPos.utm_x << "    " << curPos.utm_y << "    " 
			           << ((*lit).second) << std::endl;
		}

		pathStream.close();

		return true;
	}/*}}}*/

	bool Navigator::saveData(const std::string& outfile, const std::string& toutfile, const std::string& quadTreeFile)
	{/*{{{*/
		if( navMap == NULL ) { return false; }
		// Saving the internal geodata
		navMap->saveData(outfile);
		// Saving maps : Traversability, Potential, Utility
		std::string traversabilityFile = toutfile;
		traversabilityFile.append(".traversability.ppm");
		navMap->saveTraversability(traversabilityFile);
		std::string potentialFile = toutfile;
		potentialFile.append(".potential.ppm");
		navMap->savePotential(potentialFile);
		std::string utilityFile = toutfile;
		utilityFile.append(".utility.ppm");
	 	navMap->saveUtility(utilityFile);

		// Saving the Quadtree
		if (navQuadtree == NULL) { return false; }	
		std::ofstream quadTreeStream (quadTreeFile.c_str());
		quadTreeStream << (*navQuadtree);
		quadTreeStream.close();

		return true;
	}/*}}}*/

	bool Navigator::fastFindPath(const jafar::lgl::NavLocation& locstart, const jafar::lgl::NavLocation& locend, jafar::lgl::NavPath& navpath)
	{//{{{
		return navGraph->fastFindPath(locstart, locend, navpath);
	}//}}}

	bool Navigator::findPath(const jafar::lgl::NavLocation& locstart, const jafar::lgl::NavLocation& locend, jafar::lgl::NavPath& navpath, double& pathCost)
	{//{{{
		return navGraph->findPath(locstart, locend, navpath, pathCost);
	}//}}}

	bool Navigator::findPath(const jafar::lgl::NavLocation& locstart, const jafar::lgl::NavLocation& locend, jafar::lgl::NavPath& navpath)
	{//{{{
		return navGraph->findPath(locstart, locend, navpath);
	}//}}}

	bool Navigator::findPathOnUtility(const jafar::lgl::NavLocation& locstart, const jafar::lgl::NavLocation& locend, jafar::lgl::NavPath& navpath, double authDist)
	{//{{{
		return navGraph->findPathOnUtility(locstart, locend, navpath, authDist);
	}//}}}

	bool Navigator::updateMap(double* bufferedData, int xsize_, int ysize_)
	{//{{{
		delete navGraph;
		delete navQuadtree;
		navQuadtree = NULL;
		setMapSize(xsize_, ysize_, zsize);
		if (navMap->setData(bufferedData, NULL, xsize, ysize, jafar::lgl::GEODATA_TFLAT_BAND))
		{
			// Recreating quadtree
			navQuadtree = new jafar::lgl::Quadtree<double, jafar::lgl::GeoData>(navMap);
			// Recreating navgraph
			navGraph = new jafar::lgl::NavGraph(navMap, navQuadtree);
			return true;
		}
		return false;
	}//}}}

	bool Navigator::updateMap( jafar::lgl::GeoData *inputGData )
	{//{{{
		if (inputGData != NULL )
		{ 
			// Switching Geodata
			if (inputGData != navMap)
			{
				delete navMap;
			}
			navMap = inputGData;
		} // else we update on the current navMap which may have changed.

		// Checking if we can continue
		if (navMap == NULL) { return false; }

		// Cleaning obsolete objects
		delete navGraph;
		delete navQuadtree;

		// Setting up Geodata parameters
		setMapSize(navMap->getxsize(), navMap->getysize(), 0);
		pixelScale = navMap->getScale();
		// Creationg the geo-decomposition
		navQuadtree = new jafar::lgl::Quadtree<double, jafar::lgl::GeoData>(navMap);
		// Creationg the navigation graph
		navGraph = new jafar::lgl::NavGraph(navMap, navQuadtree);
		return true;
	}//}}}

	bool Navigator::cropPath(double vent, double vproba, jafar::lgl::NavPath& navpath)
	{//{{{
		return (multiTool->cropPath(vent, vproba, navpath));
	}//}}}

	bool Navigator::applyTool(jafar::lgl::NavPath& srcPath, jafar::lgl::NavPath& outPath, PATH_TRANSFORMATION ptrans)
	{//{{{
		return (multiTool->applyTool(srcPath, outPath, ptrans));
	}//}}}

	void Navigator::printInfo()
	{//{{{
		if (navMap != NULL)
		{
			std::cout << "#---III---  Map              ----------------" << std::endl
			<< "#---III---  Size  x " << xsize << " y " << ysize << " z " << zsize << std::endl
			<< "#---III---  Start point: (" << startx<<", "<<starty <<", "<<startz<<")"<<std::endl 
			<< "#---III---  Goal point:  (" << goalx <<", "<<goaly  <<", "<<goalz <<")"<<std::endl;

			if (navQuadtree != NULL)
			{
				std::cout << "#---III---  Decomposition    ----------------" << std::endl
				<< "#---III---  Traversability Decomposition Type : " << navMap->getTraversabilityHomogeneousType() << std::endl 
				<< "#---III---  Traversability P-Lag: " << navMap->getTraversabilityProbabilityLag() << std::endl 
				<< "#---III---  Traversability Sigma: " << navMap->getTraversabilitySigmaFactor() << std::endl 
				<< "#---III---  Rough Terrain Decomposition Type : " << navMap->getRoughHomogeneousType() << std::endl 
				<< "#---III---  Rough Terrain  P-Lag: " << navMap->getRoughProbabilityLag() << std::endl 
				<< "#---III---  Rough Terrain  Sigma: " << navMap->getRoughSigmaFactor() << std::endl;
			}
		}

		if (navGraph != NULL)
		{
			std::cout << "#---III---  Navigation Graph ----------------" << std::endl
			<< "#---III---  " << navGraph->getNumberOfVertices()  << " number of vertices." << std::endl
			<< "#---III---  " << navGraph->getNumberOfEdges()     << " number of edges." << std::endl
			<< "#---III---  " << navGraph->getNumberOfEdges()/2   << " vertices' connexions." << std::endl;
		}

		if (multiTool != NULL)
		{
			std::cout << "#---III---  Path Tool        ----------------" << std::endl
			<< "#---III---  Entropy limit: " << multiTool->getPathMaxEntropy() << std::endl
			<< "#---III---  Probability limit: " << multiTool->getPathMaxProba() << std::endl
			<< "#---III---  Subsampling limit: " << multiTool->getSubSampler() << std::endl;
		}
	}//}}}
	//---- set stuffs

	bool Navigator::setMapOrigin(double xMapOrigin_, double yMapOrigin_, double zMapOrigin_, double pixelScale_)
	{//{{{
		xMapOrigin = xMapOrigin_;
		yMapOrigin = yMapOrigin_;
		zMapOrigin = zMapOrigin_;
		pixelScale = pixelScale_;
		if (multiTool == NULL) { return false; }
		multiTool->setScale(pixelScale);
		multiTool->setGeoXRoot(xMapOrigin);
		multiTool->setGeoYRoot(yMapOrigin);
		return true;
	}//}}}

	void Navigator::setMapSize(int  xsize_, int ysize_, int zsize_)
	{//{{{
		xsize = xsize_;
		ysize = ysize_;
		zsize = zsize_;
	}//}}}

	void Navigator::setGoal(const jafar::lgl::NavLocation& A)
	{//{{{
		setGoal(A.pos_x, A.pos_y, A.pos_z);
		setGoal(A.utm_x, A.utm_y, A.utm_z);
	}//}}}

	void Navigator::setGoal(int x, int y, int z)
	{//{{{
		goalx = x;
		goaly = y;
		goalz = z;
	}//}}}

	void Navigator::setGoal(double x, double y, double z)
	{//{{{
		goalx = x;
		goaly = y;
		goalz = z;
	}//}}}

	void Navigator::setStart(const jafar::lgl::NavLocation& A)
	{//{{{
		setStart(A.pos_x, A.pos_y, A.pos_z);
		setStart(A.utm_x, A.utm_y, A.utm_z);
	}//}}}

	void Navigator::setStart(int x, int y, int z)
	{//{{{
		startx = x;
		starty = y;
		startz = z;
	}//}}}

	void Navigator::setStart(double x, double y, double z)
	{//{{{
		startutmx = x;
		startutmy = y;
		startutmz = z;
	}//}}}

	void Navigator::setPathCropers(double pce, double pcp)
	{//{{{
		if (pce >= 0.0)
		{
			multiTool->setPathMaxEntropy(pce);
		}

		if (pcp >= 0.0)
		{
			multiTool->setPathMaxProba(pcp);
		}
	}//}}}

	void Navigator::setMaxWayPts(int maxwpts)
	{//{{{
		multiTool->setMaxWayPts(maxwpts);
	}//}}}

	void Navigator::setSigmaFactor(double sig_)
	{//{{{
		navMap->setSigmaFactor(sig_);
	}//}}}

	void Navigator::setProbabilityLag(double probaLag_)
	{//{{{
		navMap->setProbabilityLag(probaLag_);
	}//}}}

	void Navigator::setPathToolPathMaxDistance(double pmd)
	{//{{{
		multiTool->setPathMaxDistance(pmd);
	}//}}}

	void Navigator::setPathToolSubSampler(double setsbs)
	{//{{{
		multiTool->setSubSampler(setsbs);
	}//}}}

	void Navigator::setPathToolInterpolationDist(double interpDist_)
	{//{{{
		multiTool->setInterpolationDistance(interpDist_);
	}//}}}

	void Navigator::setPathToolInterpolationIndexStart(int indexstart_)
	{//{{{
		multiTool->setInterpolationStartPointIdx(indexstart_);
	}//}}}

	void Navigator::setPathToolInterpolationIndexStop(int indexstop_)
	{//{{{
		multiTool->setInterpolationStopPointIdx(indexstop_);
	}//}}}

	//---- get stuffs
	
	jafar::lgl::NavGraph::NavGraphContainer* Navigator::getGraph()
	{/*{{{*/
		if (navGraph != NULL)
		{
			return navGraph->getGraph();
		}
		return NULL;
	}/*}}}*/

	jafar::lgl::Decomp* Navigator::getDecomposition()
	{/*{{{*/
		if (navQuadtree != NULL)
		{
			return navQuadtree->getDecomp();
		}
		return NULL;
	}/*}}}*/

}} // end of namespaces
