/**
 * \file pathTool.cpp
 *
 * Source file for the rednav path tool.
 *
 * \date 2011-01-07
 * \author redouane.boumghar@laas.fr
 * \ingroup rednav
 */

#include <rednav/pathTool.hpp>

namespace jafar {
namespace rednav {

	// Constructor that does nothing but giving calls to tools
	PathTool::PathTool()
	{ // {{{
		// Default values for all parameters
		subsampler = 4.0;
		maxwaypts  = 0;
		pathMaxEntropy = 0.69135;
		pathMaxProba = 0.8;
		cellScale = 1.0;
		pathMaxDistance = 50.0;
		pathMaxStepDistance = 50.0;
		pathMaxDevDistance = 1.50;
		sensorRadius = 12.0;
		interpolationDistance =  1.0;
		interpolationStartPointIdx = 0;
		interpolationStopPointIdx = 1;
		navgraph = NULL;
	} // }}}

	// Tooled Constructor
	PathTool::PathTool(jafar::lgl::NavPath& path, PATH_TRANSFORMATION trsf, jafar::lgl::NavGraph* ng_)
	{ // {{{
		subsampler = 4.0;
		maxwaypts  = 0;
		pathMaxEntropy = 0.69135;
		pathMaxProba = 0.8;
		cellScale = 1.0;
		geoXRoot = 0.0;
		geoYRoot = 0.0;
		pathMaxDistance = 50.0;
		pathMaxStepDistance = 50.0;
		pathMaxDevDistance = 1.50;
		sensorRadius = 12.0;
		interpolationDistance =  1.0;
		interpolationStartPointIdx = 0;
		interpolationStopPointIdx = 1;
		navgraph = ng_;
	} // }}}

	//-------------------------- Destructor ---------
	PathTool::~PathTool()
	{ // {{{
	} // }}}

	//-------------------------- Setting Functions  ---------
	void PathTool::setMaxWayPts( int n )
	{//{{{
		maxwaypts = n;
	}//}}}

	void PathTool::setPathMaxEntropy( double pme )
	{//{{{
		pathMaxEntropy = pme;
	}//}}}

	void PathTool::setPathMaxProba( double pmp )
	{//{{{
		pathMaxProba = pmp;
	}//}}}

	void PathTool::setSubSampler( double setsb)
	{//{{{
		subsampler = setsb;
	}//}}}

	void PathTool::setPathMaxDistance(double pmd)
	{//{{{
		pathMaxDistance = pmd;
	}//}}}

	void PathTool::setPathMaxStepDistance(double pmd)
	{//{{{
		pathMaxStepDistance = pmd;
	}//}}}

	void PathTool::setPathMaxDevDistance(double pmd)
	{//{{{
		pathMaxDevDistance = pmd;
	}//}}}

	void PathTool::setSensorRadius(double ssr)
	{//{{{
		sensorRadius = ssr;
	}//}}}

	void PathTool::setScale( double cellres)
	{//{{{
		cellScale = cellres;
	}//}}}

	void PathTool::setGeoXRoot( double geoxr_)
	{//{{{
		geoXRoot = geoxr_;
	}//}}}

	void PathTool::setGeoYRoot( double geoyr_)
	{//{{{
		geoYRoot = geoyr_;
	}//}}}

	void PathTool::setInterpolationStartPointIdx(int startidx_)
	{//{{{
		interpolationStartPointIdx = startidx_;
	}//}}}

	void PathTool::setInterpolationStopPointIdx(int stopidx_)
	{//{{{
		interpolationStopPointIdx = stopidx_;
	}//}}}

	void PathTool::setInterpolationDistance(double interpDist_)
	{//{{{
		interpolationDistance = interpDist_;
	}//}}}

	void PathTool::setNavGraph(jafar::lgl::NavGraph* ng_)
	{//{{{
		navgraph = ng_;
	}//}}}

	//-------------------------- Getting Functions  ---------
	double PathTool::getPathMaxEntropy()
	{//{{{
		return pathMaxEntropy;
	}//}}}

	double PathTool::getPathMaxProba()
	{//{{{
		return pathMaxProba;
	}//}}}

	double PathTool::getSubSampler()
	{//{{{
		return subsampler;
	}//}}}

	double PathTool::getPathMaxDistance()
	{//{{{
		return pathMaxDistance;
	}//}}}

	double PathTool::getPathMaxStepDistance()
	{//{{{
		return pathMaxStepDistance;
	}//}}}

	double PathTool::getPathMaxDevDistance()
	{//{{{
		return pathMaxDevDistance;
	}//}}}

	double PathTool::getSensorRadius()
	{//{{{
		return sensorRadius;
	}//}}}

	double PathTool::getScale()
	{//{{{
		return cellScale;
	}//}}}

	double PathTool::getGeoXRoot()
	{//{{{
		return geoXRoot;
	}//}}}

	double PathTool::getGeoYRoot()
	{//{{{
		return geoYRoot;
	}//}}}

	int PathTool::getInterpolationStartPointIdx()
	{//{{{
		return interpolationStartPointIdx;
	}//}}}

	int PathTool::getInterpolationStopPointIdx()
	{//{{{
		return interpolationStopPointIdx;
	}//}}}
	//-------------------------- Small utilities Functions  ----
	double PathTool::navLocToMeters(int a)
	{ //{{{
		return (cellScale *(((double)a)/2.0));
	} //}}}

	double PathTool::navLocDistance(const jafar::lgl::NavLocation& A, const jafar::lgl::NavLocation& B)
	{ //{{{
		double Ax, Ay;
		double Bx, By;
		Ax = navLocToMeters(A.pos_x);
		Ay = navLocToMeters(A.pos_y);
		Bx = navLocToMeters(B.pos_x);
		By = navLocToMeters(B.pos_y);
		return (sqrt(
				  (Ax-Bx)*(Ax-Bx)
				+ (Ay-By)*(Ay-By) ));
	} //}}}

	double PathTool::getOrthogonalHeight(const jafar::lgl::NavLocation& A, const jafar::lgl::NavLocation& B, const jafar::lgl::NavLocation& C)
	{ //{{{
		double mslope = 0.0;
		double mslope_c = 0.0; // useful coefficient
		double dx, dy; // vector AB - AB slope
		double cx, cy; // projected point relative to A coordinates.
		double hx = 0.0, hy = 0.0; // projected point coordinates.
		double oheight = 0.0;

		// Getting slope of the base segment [AB]
		dx = (double)(B.pos_x - A.pos_x);
		dy = (double)(B.pos_y - A.pos_y);
		// Point coordinates that we want to othorgonal-project on AB segment of.
		cx = (double)(C.pos_x - A.pos_x);
		cy = (double)(C.pos_y - A.pos_y);
		if ( dx != 0.0)
		{
			mslope = dy / dx;
			mslope_c = dx + mslope * dy;
			if (mslope_c != 0.0)
			{
				hx = (cx * dx + cy * dy) / mslope_c;
			} else {
				// should never happen as its the inverse definition of mslope.
			}
			hy = mslope * hx;
			hx = hx + (double)A.pos_x;
			hy = hy + (double)A.pos_y;
		} else {
			// Vertical case
			hx = (double)A.pos_x; 
			hy = (double)C.pos_y; 
		}

		cx = (double)(C.pos_x);
		cy = (double)(C.pos_y);
		oheight = sqrt( (hx-cx)*(hx-cx) + (hy-cy)*(hy-cy) );
		oheight = (cellScale * ((oheight)/2.0) );
		return oheight;
	} //}}}

	bool PathTool::surroundingsDanger(const jafar::lgl::NavLocation& A, float& fodist)
	{ //{{{
		double tmpdist  = 0.0;
		bool notChecked = true;
		jafar::lgl::DecompCluster* dcluster;
		// Limit distance radius under which it is not safe to have obstacle
		float limdist = 3.0*pathMaxDevDistance;
		float asaferadius = limdist;

		// Get anchor vertex in graph
		jafar::lgl::NavLocation graphPoint = (navgraph->getGraph())->getClosestLocation(A);
		jafar::lgl::NavGraph::NavGraphContainer::Vertex avertex;
		(navgraph->getGraph())->getVertex(graphPoint, avertex, false);

		// Init Iteration over neighbor hood in graph
		jafar::lgl::NavGraph::NavGraphContainer::GraphContainer* graphee =
			(navgraph->getGraph())->getGCtn();
		jafar::lgl::NavGraph::NavGraphContainer::out_edge_iterator ei, edge_end;
		jafar::lgl::NavGraph::NavGraphContainer::Vertex ctarget;
		jafar::lgl::NavGraph::NavGraphContainer::out_edge_iterator ei2, edge_end2;
		jafar::lgl::NavGraph::NavGraphContainer::Vertex ctarget2;
		std::list<jafar::lgl::NavGraph::NavGraphContainer::Vertex> checkedVertex;
		std::list<jafar::lgl::NavGraph::NavGraphContainer::Vertex> openVertex;
		std::list<jafar::lgl::NavGraph::NavGraphContainer::Vertex>::iterator vi;

		openVertex.push_back(avertex);
		while (openVertex.size() > 0)
		{
			avertex = openVertex.front();
			openVertex.pop_front();
			for (tie(ei,edge_end) = out_edges(avertex, *graphee); ei != edge_end; ++ei)
			{
				ctarget = target((*ei), *graphee);
				tmpdist = navLocDistance(graphPoint, ((*graphee)[ctarget]).pos);
				// Threat this vertex if not to far from our central graphPoint
				if(tmpdist < asaferadius)
				{
					// Check this vertex is not already open or checked
					notChecked = true;
					for(vi = openVertex.begin(); vi != openVertex.end(); ++vi)
					{
						if((*vi) == ctarget)
						{
							notChecked = false;
							break;
						}
					}

					if (notChecked)
					for(vi = checkedVertex.begin(); vi != checkedVertex.end(); ++vi)
					{
						if((*vi) == ctarget)
						{
							notChecked = false;
							break;
						}
					}

					// Checking distance and corresponding edge if not already checked.
					if (notChecked)
					{
						// CTarget edge at ei
						(navgraph->getPositionManager())->getCluster(
							((*graphee)[(*ei)]).cluster,   &dcluster  );
						if (    ((dcluster->getAttributes())->infotropy[jafar::lgl::T_SYMBOL] > 0.40)
							||  ((dcluster->getAttributes())->certainty[jafar::lgl::T_OBSTACLE] > 0.5) )
						{
							asaferadius = tmpdist;
						}

					}
					// Add vertex
					checkedVertex.push_back(ctarget);
				}

				// Add close neighbors of ctarget to the open list
				for (tie(ei2,edge_end2) = out_edges(ctarget, *graphee); ei2 != edge_end2; ++ei2)
				{
					ctarget2 = target((*ei2), *graphee);
					// Check this vertex is not already open or checked
					notChecked = true;
					for(vi = openVertex.begin(); vi != openVertex.end(); ++vi)
					{
						if((*vi) == ctarget2)
						{
							notChecked = false;
							break;
						}
					}

					if (notChecked)
					for(vi = checkedVertex.begin(); vi != checkedVertex.end(); ++vi)
					{
						if((*vi) == ctarget2)
						{
							notChecked = false;
							break;
						}
					}

					// Add this vertex to its corresponding list
					tmpdist = navLocDistance(graphPoint, ((*graphee)[ctarget2]).pos);
					if(notChecked && tmpdist < asaferadius)
					{
						openVertex.push_back(ctarget2);
					} else {
						checkedVertex.push_back(ctarget2);
					}
				}



			}
		}

		if (asaferadius < limdist)
		{
			return true;
		}
		return false;
	} //}}}

	//-------------------------- Other Big Functions  ---------
	bool PathTool::applyTool(const jafar::lgl::NavPath& inpath, jafar::lgl::NavPath& outpath, PATH_TRANSFORMATION ptrans)
	{//{{{
		switch(ptrans)
		{
			case PATH_TWO_POINTS:
				return (twoPoints(inpath, outpath));
				break;
			case PATH_RDV_CROP:
				return (cropPath(inpath, outpath));
			case PATH_TWO_POINTS_LIMIT:
				return (twoPointsFromCurve(inpath, outpath));
			case PATH_DISTANCE_LIMIT:
				return (cropDistance(inpath, outpath, pathMaxDistance));
			case PATH_INTERPOLATION:
				return (pathInterpolation(inpath, outpath, interpolationDistance, interpolationStartPointIdx, interpolationStopPointIdx));
			case PATH_SAFESAMPLING:
				return (pathSafeSampling(inpath, outpath));
			case PATH_SMARTSAMPLING:
				return (pathSmartSampling(inpath, outpath));
			default: // PATH_COMMON_FILTER
				return (subSample(inpath, outpath, subsampler));
		}
		return false;
	}//}}}

	bool PathTool::cropDistance(const jafar::lgl::NavPath& foundPath, jafar::lgl::NavPath& tPath, double distLimit)
	{//{{{
		// Init.
		jafar::lgl::NavPath::const_iterator path_it;

		double distTrack = 0.0;
		double cox  = 0.0;
		double coy  = 0.0;
		double pcox = 0.0;
		double pcoy = 0.0;

		// Temporary pointers to the start point
		path_it = foundPath.begin();
		if (path_it == foundPath.end()) { return false; }
		pcox = navLocToMeters(((*path_it).first).pos_x) ;
		pcoy = navLocToMeters(((*path_it).first).pos_y) ;

		tPath.clear();

		//---------- Prepare two-points ways for Ressac v1
		for( path_it  = foundPath.begin();
			 path_it != foundPath.end();
			 ++path_it)
		{
			cox = navLocToMeters(((*path_it).first).pos_x) ;
			coy = navLocToMeters(((*path_it).first).pos_y) ;

			distTrack += sqrt((pcox-cox) * (pcox-cox)
			                + (pcoy-coy) * (pcoy-coy));

			// Adding this node to the track
			tPath.push_back((*path_it));

			// Check if difference is not too big else stop here.
			if ( (distTrack ) >= distLimit)
			{
				break;
			}

			pcox = cox;
			pcoy = coy;
		}

		return true;
	}//}}}

	bool PathTool::twoPointsFromCurve(const jafar::lgl::NavPath& foundPath, jafar::lgl::NavPath& tPath)
	{//{{{
		// Init.
		jafar::lgl::NavPath::const_iterator path_start;
		jafar::lgl::NavPath::const_iterator path_it;

		double distBirdTrack = 0.0;
		double distTrack = 0.0;
		double startx  = 0.0;
		double starty  = 0.0;
		double cox  = 0.0;
		double coy  = 0.0;
		double pcox = 0.0;
		double pcoy = 0.0;
		// Temporary pointers to the start point
		path_start = foundPath.begin();
		if (path_start == foundPath.end()) { return false; }
		startx = navLocToMeters(((*path_start).first).pos_x) ;
		starty = navLocToMeters(((*path_start).first).pos_y) ;
		pcox = startx;
		pcoy = starty;

		tPath.clear();
		tPath.push_back((*path_start));

		//---------- Prepare two-points ways for Ressac v1
		int Ndone = 0;
		for( path_it  = foundPath.begin();
			 path_it != foundPath.end();
			 ++path_it)
		{
			cox = navLocToMeters(((*path_it).first).pos_x) ;
			coy = navLocToMeters(((*path_it).first).pos_y) ;

			distTrack += sqrt((pcox-cox) * (pcox-cox)
			                + (pcoy-coy) * (pcoy-coy));

			distBirdTrack = sqrt((startx-cox) * (startx-cox)
			                   + (starty-coy) * (starty-coy));

			// Check if difference is not too big else stop here.
			if (   (distTrack - distBirdTrack) > sensorRadius
				|| (distBirdTrack >= pathMaxDistance))
			{
				Ndone = 2;
				tPath.push_back((*path_it));
				break;
			}

			pcox = cox;
			pcoy = coy;
		}

		//----- Filling a 2-waypoints path
		if (Ndone != 2)
		{
			path_it = foundPath.end();
			if (path_start != path_it)
			{
				// Case with a quasi-straight path shorter than pathMaxDistance
				// straight line from Start To End
				--path_it;
				tPath.push_back((*path_it));
			} else {
				// Two way points on the same point
				// in case of too short raw paths.
				tPath.push_back((*path_start));
				return false;
			}
		}

		return true;
	}//}}}

	bool PathTool::twoPoints(const jafar::lgl::NavPath& foundPath, jafar::lgl::NavPath& tPath)
	{//{{{
		jafar::lgl::Attributes emptyAtt;
		// Init.
		jafar::lgl::NavPath::const_iterator path_it;
		jafar::lgl::NavPath::const_iterator path_it2;

		// Temporary pointers to the start and goal points
		path_it = foundPath.begin();
		if (path_it == foundPath.end()) { return false; }
		path_it2 = foundPath.end(); --path_it2;
		if (path_it2 == foundPath.end()) { return false; }

		//---------- Prepare two-points ways for Ressac v1
		// Ressac v1 : ressac only able to follow two points for mapping
		//----- Find the farthest points to the line start-goal on each side
		// start-goal vector coordinates
		double vx_sg = (double)(((*path_it2).first).pos_x - ((*path_it).first).pos_x);
		double vy_sg = (double)(((*path_it2).first).pos_y - ((*path_it).first).pos_y);
		// 'Absolute' start coordinates
		double co_sx = (double)((*path_it).first).pos_x;
		double co_sy = (double)((*path_it).first).pos_y;
		// Relative Goal coordinates with transformation Start --> (0,0)
		double co_gx = (double)((*path_it2).first).pos_x - co_sx;
		double co_gy = (double)((*path_it2).first).pos_y - co_sy;
		// Considered path point coordinates
		double co_ax = 0.0;
		double co_ay = 0.0;
		// 'Absolute' On-DirectLine projected points coordinates
		double co_hx = 0.0;
		double co_hy = 0.0;
		// Slope coefficient of the direct line between start and goal points
		bool mslope_vertical = false;
		double mslope   = 0.0;
		double mslope_c = 0.0;
		if (vx_sg != 0.0)
		{
			mslope = vy_sg / vx_sg;
			mslope_c = co_gx + mslope * co_gy;
		} else {
			mslope_vertical =  true;
		}
		// Distance parameters
		double dist = 0.0;
		double maxdist_a = 0.0;
		double maxdist_b = 0.0;
		// Axis relative signed position
		double sign_a = 0.0;

		// To tell if we have found at least one point on the positive side
		int plus_ni  = -1;
		double plus_ax = 0.0;
		double plus_ay = 0.0;
		// To tell if we have found at least one point on the negative side
		int minus_ni = -1;
		double minus_ax = 0.0;
		double minus_ay = 0.0;

		int N = 0;
		for( path_it  = foundPath.begin();
			 path_it != foundPath.end();
			 ++path_it)
		{
			co_ax = (double)((*path_it).first).pos_x ;
			co_ay = (double)((*path_it).first).pos_y ;
			if (mslope_vertical)
			{
				co_hx = co_sx;
				co_hy = co_ay;
			} else {
				// Getting relative position when start --> (0,0)
				co_ax = co_ax - co_sx;
				co_ay = co_ay - co_sy;
				// Calculating relative projection
				co_hx = (co_ax * co_gx + co_ay * co_gy) / mslope_c;
				co_hy = mslope * co_hx;
				// Translating to the main frame
				co_ax = co_ax + co_sx;
				co_ay = co_ay + co_sy;
				co_hx = co_hx + co_sx;
				co_hy = co_hy + co_sy;
			}

			dist = sqrt( (co_hx-co_ax) * (co_hx-co_ax)
					   + (co_hy-co_ay) * (co_hy-co_ay) ) ;
			// Saving farthest point on its side
			sign_a = (co_hx - co_ax)*(co_gy)
				   + (co_hy - co_ay)*(co_gx);
			if ( sign_a > 0.0 )
			{
				if (dist > maxdist_a)
				{
					maxdist_a = dist;
					plus_ax = co_ax;
					plus_ay = co_ay;
					plus_ni = N;
				}
			} else {
				if (dist >= maxdist_b)
				{
					maxdist_b = dist;
					minus_ax = co_ax;
					minus_ay = co_ay;
					minus_ni = N;
				}
			}

			// Full path index count
			N++;
		}

		//----- Filling a 2-waypoints path
		tPath.clear();
		std::pair<jafar::lgl::NavLocation, jafar::lgl::Attributes> insertor;
		insertor.second = emptyAtt;
		if ( plus_ni >= 0 && minus_ni >= 0)
		{
			// There is a point on each side of the start-goal line
			insertor.first = jafar::lgl::NavLocation(plus_ax, plus_ay);
			tPath.push_back(insertor);
			insertor.first = jafar::lgl::NavLocation(minus_ax, minus_ay);
			tPath.push_back(insertor);
		} else {
			if (plus_ni >=0 && minus_ni == -1)
			{
				insertor.first = jafar::lgl::NavLocation(plus_ax, plus_ay);
				tPath.push_back(insertor);
				path_it = foundPath.begin();
				if (path_it != foundPath.end())
				{
					tPath.push_back((*path_it));
				} else {
					return false;
				}
			} else {
				if (minus_ni >=0 && plus_ni == -1)
				{
					insertor.first = jafar::lgl::NavLocation(minus_ax, minus_ay);
					tPath.push_back(insertor);
					path_it = foundPath.begin();
					if (path_it != foundPath.end())
					{
						tPath.push_back((*path_it));
					} else {
						return false;
					}
				} else {
					// No points found ? quitely impossible :)
					return false;
				}
			}
		}

		return true;
	}//}}}

	bool PathTool::subSample(const jafar::lgl::NavPath& foundPath, jafar::lgl::NavPath& tPath, double subsamp)
	{//{{{
		//------------ variables
		double x_new = 0.0;
		double y_new = 0.0;
		double x_prevpos = 0.0;
		double y_prevpos = 0.0;
		double curvDist = 0.0;
		int N = -1;
		int Nend = (int)(foundPath.size());
		bool lastPointAdded = false;
		jafar::lgl::NavPath::const_iterator path_it;
		tPath.clear();

		//--- Subsample the input path to give something smoother for the
		//--- obstacle avoider
		for( path_it = foundPath.begin();
			 path_it != foundPath.end();
			 ++path_it)
		{
			// Count to check if we are adding the last point
			Nend--;
			// Correcting coordinates in the superimposed LGL frame
			x_new = (double)((*path_it).first).pos_x;
			x_new = (x_new/2.0)*cellScale;
			y_new = (double)((*path_it).first).pos_y;
			y_new = (y_new/2.0)*cellScale;
			if (N==-1)
			{
				// Not using first point of the trajectory as it is the
				// position of the vehicule itself or the first point in the
				// trajectory. Our reference.
				N++;
				//tPath.push_back((*path_it));
			} else {
				// Resampling trajectory to a more loose one
				x_prevpos = x_prevpos - x_new;
				x_prevpos = x_prevpos*x_prevpos;
				y_prevpos = y_prevpos - y_new;
				y_prevpos = y_prevpos*y_prevpos;
				curvDist = curvDist + sqrt(x_prevpos + y_prevpos);
				if ( curvDist >= subsamp)
				{
					// Adding points to the transformed path
					tPath.push_back((*path_it));
					curvDist = 0.0;
					N++;
					// Checking if we just added the last point
					if (Nend == 0)
					{
						lastPointAdded = true;
					}
				} // else not adding it.
			}

			if ( (maxwaypts > 0) && (N > maxwaypts) )
			{
				// No more waypoints needed.
				return true;
			}
			x_prevpos = x_new;
			y_prevpos = y_new;
		}

		// Adding last trajectory point if not already in.
		if (!lastPointAdded)
		{
			path_it = foundPath.end(); --path_it;
			if (path_it == foundPath.end()) { return false; }
			tPath.push_back((*path_it));
		}

		if (N <= 0)
		{
			return false;
		}
		return true;
	}//}}}

	bool PathTool::cropPath(const jafar::lgl::NavPath& srcPath, jafar::lgl::NavPath& outPath)
	{//{{{
		if ( pathMaxEntropy <= 0.0 && pathMaxProba <= 0.0)
		{
			// No crop possible.
			return false;
		}
		jafar::lgl::NavPath::const_iterator path_it;
		int n = 0;
		outPath.clear();
		std::pair<jafar::lgl::NavLocation, jafar::lgl::Attributes> insertor;
		for(path_it = srcPath.begin(); (path_it != srcPath.end()); ++path_it)
		{
			// Testing if limits (pathMaxEntropy and pathMaxProba) are reach OR if maxwaypts is
			// set to more than zero crop the path to that maximum allowed
			// number of waypoints.
			if (   (pathMaxEntropy > 0.0 && (((*path_it).second).infotropy[jafar::lgl::T_SYMBOL])   >= pathMaxEntropy)
				|| (pathMaxProba   > 0.0 && (((*path_it).second).certainty[jafar::lgl::T_OBSTACLE]) >= pathMaxProba)
				|| (maxwaypts > 0 && n > maxwaypts) )
			{
				// Working on copy
				//srcPath.erase(path_it, srcPath.end());
				return true;
			}
			insertor.first  = path_it->first;
			insertor.second = path_it->second;
			outPath.push_back(insertor);
			//outPath.push_back(std::make_pair(path_it->first, path_it->second));
			n++;
		}
		// No cropping has been done
		return false;
	}//}}}

	bool PathTool::cropPath(double vent, double vproba, jafar::lgl::NavPath& navpath)
	{//{{{
		if ( vent <= 0.0 && vproba <= 0.0)
		{
			// No crop possible.
			return false;
		}
		jafar::lgl::NavPath::iterator path_it;
		int n = 0;
		for(path_it = navpath.begin(); (path_it != navpath.end()); ++path_it)
		{
			// Testing if limits (vent and vproba) are reach OR if maxwaypts is
			// set to more than zero crop the path to that maximum allowed
			// number of waypoints.
			if (   (vent   > 0.0 && (((*path_it).second).infotropy[jafar::lgl::T_SYMBOL])   >= vent )
				|| (vproba > 0.0 && (((*path_it).second).certainty[jafar::lgl::T_OBSTACLE]) >= vproba)
				|| (maxwaypts > 0 && n > maxwaypts) )
			{
				navpath.erase(path_it, navpath.end());
				return true;
			}
			n++;
		}
		// No cropping has been done
		return false;
	}//}}}

	bool PathTool::pathInterpolation(const jafar::lgl::NavPath& srcPath, jafar::lgl::NavPath& tPath, double smallestInterval, int masterP1, int masterP2)
	{//{{{
		//------------ variables
		double x_cur = 0.0;
		double y_cur = 0.0;
		int N = 0;
		int Ntot = 0;
		double cosAngle = 0.0;
		double sinAngle = 0.0;
		// Interpolation start point coordinates
		double Ax = 0.0;
		double Ay = 0.0;
		// Interpolation end point coordinates
		double Bx = 0.0;
		double By = 0.0;
		double ABdist = 0.0;
		double curvDist = smallestInterval;
		std::pair<jafar::lgl::NavLocation, jafar::lgl::Attributes> insertor;
		jafar::lgl::NavPath::const_iterator path_it;
		//------------ end of variables

		// Clearing resulting path
		tPath.clear();

		//--- Checking consistency of indexes:
		// Doing interpolation on the first segment if masterP1 and masterP2 < 0
		if (masterP1 < 0)
		{ masterP1 = 0; }
		if (masterP2 < 0 || masterP2 <= masterP1)
		{ masterP2 = masterP1 + 1; }

		//--- Interpolate only between masterP1-th and masterP2-th
		//--- but browse all points to let them in the resulting path
		for( path_it = srcPath.begin();
			 path_it != srcPath.end();
			 ++path_it)
		{
			// Correcting coordinates in the superimposed LGL frame
			x_cur = (double)((*path_it).first).pos_x;
			x_cur = (x_cur/2.0)*cellScale;
			y_cur = (double)((*path_it).first).pos_y;
			y_cur = (y_cur/2.0)*cellScale;

			if (N==masterP1)
			{ //{{{
				Ax = x_cur;
				Ay = y_cur;
				if (N != 0)
				{
					// Not using first point of the trajectory as it is the
					// position of the vehicule itself or the first point in the
					// trajectory. Our reference.
					tPath.push_back((*path_it));
					Ntot++;
				}
			} else {

				// Interpolate if before interpolation end point.
				if ( N <= masterP2)
				{//{{{
					// Point A is an init point or a previous-point memory
					// We just have to set point B.
					Bx = x_cur;
					By = y_cur;

					// Getting cosinus and sinus
					ABdist = sqrt((Bx-Ax)*(Bx-Ax) + (By-Ay)*(By-Ay));
					if (ABdist > 0.0)
					{
						cosAngle = (Bx - Ax) / ABdist;
						sinAngle = (By - Ay) / ABdist;
					} else {
						cosAngle = 0.0;
						sinAngle = 0.0;
					}
					// Copying current attributes. They'll be the same for all
					// following interpolated points.
					insertor.second = (*path_it).second;

					//--- Interpolating until we can not
					while ( curvDist < ABdist)
					{//{{{
						// Getting the position variation from A to the next
						// interpolated point.
						x_cur = Ax + curvDist * cosAngle;
						y_cur = Ay + curvDist * sinAngle;

						// Sticking to the grid
						insertor.first = jafar::lgl::NavLocation(
							(int)round(x_cur*2.0/cellScale),
							(int)round(y_cur*2.0/cellScale)
						);
						(insertor.first).updateUTMPos(geoXRoot, geoYRoot, cellScale, cellScale);
						// Adding this interpolated point
						tPath.push_back(insertor);
						Ntot++;
						// Going to next trajectory interpolation point
						curvDist = curvDist + smallestInterval;
					}//}}}
					// Resetting curve distance.
					curvDist = smallestInterval;
					// Restarting from B.
					Ax = Bx;
					Ay = By;
				}//}}}

			}//}}}

			// Interpolation is over or not done and adding next master point
			tPath.push_back((*path_it));
			// Updating counters
			N++;
			Ntot++;

			if ( (maxwaypts > 0) && (Ntot > maxwaypts) )
			{
				// No more waypoints needed.
				return true;
			}

		}

		// In case there is no point in the paths.
		if (Ntot == 0) { return false; }

		return true;
	}//}}}

	bool PathTool::pathSafeSampling(const jafar::lgl::NavPath& foundPath, jafar::lgl::NavPath& tPath)
	{//{{{
		// Init.
		jafar::lgl::NavPath::const_iterator path_it;
		jafar::lgl::NavPath::const_iterator intpath_it;

		// Path pointers
		jafar::lgl::NavLocation baseLoc;
		jafar::lgl::NavLocation curLoc;
		jafar::lgl::NavVertex prevPos;
		jafar::lgl::NavVertex farPos;

		double tmp = 0.0;

		// Path length - limitating
		double distTrack = 0.0;
		// Hauteur - triangle Height - limitating
		double distMaxToTrack = 0.0;
		// Temporary pointers to the start point
		path_it = foundPath.begin();
		if (path_it == foundPath.end()) { return false; }
		// Setting start previous loc for loop
		baseLoc = (*path_it).first;
		prevPos = (*path_it);
		farPos = (*path_it);

		// Cleaning output path
		tPath.clear();
		tPath.push_back((*path_it));

		jafar::lgl::NavPath intermediaryPoints;

		//---------- Low-Height Triangle sampling
		path_it = foundPath.begin() ;
		++path_it;
		int NVertex = 1;
		int NVertexTotal = foundPath.size();
		for( ; path_it != foundPath.end();
			 ++path_it)
		{
			NVertex++;
			curLoc = (*path_it).first;

			// Tracking path length on the baseLoc-curLoc portion
			distTrack = distTrack + navLocDistance(curLoc, prevPos.first);

			// Tracking the farthest point to the base<->cur segment
			intermediaryPoints.push_front(prevPos);
			for (intpath_it = intermediaryPoints.begin();
			     intpath_it != intermediaryPoints.end();
				 ++intpath_it)
			{
				tmp = getOrthogonalHeight(baseLoc, curLoc, (*intpath_it).first);
				if (tmp > distMaxToTrack)
				{
					distMaxToTrack = tmp;
					farPos = (*intpath_it);
				}
			}

			// Add points and restart loop if differences are too big,
			// Or if we are at the last point in which case we had it anyway.
			if (   (distTrack >= pathMaxStepDistance)
				|| (distMaxToTrack >= pathMaxDevDistance)
				|| NVertex == NVertexTotal )
			{
				// ADDING most deviated point
				if (distMaxToTrack >= pathMaxDevDistance)
				{
					tPath.push_back(farPos);
				}
				// ADDING current point (next base point)
				tPath.push_back((*path_it));
				// Setting params to restart resampling
				baseLoc = (*path_it).first;
				distTrack = 0.0;
				distMaxToTrack = 0.0;
				intermediaryPoints.clear();
			}

			prevPos = (*path_it);
		}

		return true;
	}//}}}

	bool PathTool::pathSmartSampling(const jafar::lgl::NavPath& foundPath, jafar::lgl::NavPath& tPath)
	{//{{{

		if (navgraph == NULL)
		{
			// Nothing can be done without a navgraph
			tPath = foundPath;
			return false;
		}

		// Init.
		jafar::lgl::NavPath::const_iterator path_it;
		jafar::lgl::NavPath::const_iterator intpath_it;
		std::pair<jafar::lgl::NavLocation, jafar::lgl::Attributes> insertor;

		// Path pointers
		jafar::lgl::NavLocation baseLoc;
		jafar::lgl::NavLocation curLoc;
		jafar::lgl::NavVertex prevPos;
		jafar::lgl::NavVertex farPos;

		double tmp = 0.0;
		float ftmp = 0.0;

		// Path length - limitating
		double distTrack = 0.0;
		// Hauteur - triangle Height - limitating
		double distMaxToTrack = 0.0;
		// Temporary pointers to the start point
		path_it = foundPath.begin();
		if (path_it == foundPath.end()) { return false; }
		// Setting start previous loc for loop
		baseLoc = (*path_it).first;
		prevPos = (*path_it);
		farPos = (*path_it);

		// Cleaning output path
		tPath.clear();
		tPath.push_back((*path_it));

		jafar::lgl::NavPath intermediaryPoints;

		//---------- Low-Height Triangle sampling
		path_it = foundPath.begin() ;
		++path_it;
		int NVertex = 1;
		int NVertexTotal = foundPath.size();
		for( ; path_it != foundPath.end();
			 ++path_it)
		{
			NVertex++;
			curLoc = (*path_it).first;

			// Tracking path length on the baseLoc-curLoc portion
			distTrack = distTrack + navLocDistance(curLoc, prevPos.first);

			// Tracking the farthest point to the base<->cur segment
			intermediaryPoints.push_front(prevPos);
			for (intpath_it = intermediaryPoints.begin();
			     intpath_it != intermediaryPoints.end();
				 ++intpath_it)
			{
				tmp = getOrthogonalHeight(baseLoc, curLoc, (*intpath_it).first);
				if (tmp > distMaxToTrack)
				{
					distMaxToTrack = tmp;
					farPos = (*intpath_it);
				}
			}

			// Add points and restart loop if differences are too big,
			// Or if we are at the last point in which case we had it anyway.
			if (   (distTrack >= pathMaxStepDistance)
				|| (distMaxToTrack >= pathMaxDevDistance)
				|| NVertex == NVertexTotal )
			{
				// ADDING most deviated point
				if (distMaxToTrack >= pathMaxDevDistance)
				{
					tPath.push_back(farPos);
				}
				// ADDING current point (next base point)
				tPath.push_back((*path_it));
				// Setting params to restart resampling
				baseLoc = (*path_it).first;
				distTrack = 0.0;
				distMaxToTrack = 0.0;
				intermediaryPoints.clear();
			} else {
				if ( surroundingsDanger(((*path_it).first), ftmp) )
				{
					insertor.first  = (*path_it).first;
					insertor.second = (*path_it).second;
					insertor.first.setSafeRadius(ftmp);
					// ADDING current point (next base point)
					tPath.push_back(insertor);
					// Setting params to restart resampling
					baseLoc = (*path_it).first;
					distTrack = 0.0;
					distMaxToTrack = 0.0;
					intermediaryPoints.clear();
				}
			}

			prevPos = (*path_it);
		}

		return true;
	}//}}}


}} // end of namespaces
