/**
 * \file sensorModel.cpp
 *
 * Source file defining the probabilistic sensor model
 *
 * \author Redouane Boumghar
 *
 * \ingroup rednav
 *
 **/

#include <rednav/sensorModel.hpp>


namespace jafar {
namespace rednav {

using namespace jafar::lgl;

SensorModel::SensorModel():
	sensormat(NULL),
	sensormatSize(0),
	randomModel(0),
	sensorRange(0.0),
	maxChanfreinDist(0.0),
	shortRangeChanfreinDist(3.0),
	distanceDampingModel(2),
	distanceDampingFactor(1.2),
	visibilityMinRadius(2.0)
{ //{{{
} //}}}

SensorModel::SensorModel(int matSize):
	sensormatSize(matSize),
	randomModel(0),
	sensorRange(0.0),
	maxChanfreinDist(0.0),
	shortRangeChanfreinDist(3.0),
	distanceDampingModel(2),
	distanceDampingFactor(1.2),
	visibilityMinRadius(2.0)
{ //{{{
	sensormat = (double*)calloc(matSize*matSize, sizeof(double));
} //}}}

SensorModel::~SensorModel() 
{ //{{{
	free(sensormat);
} //}}}

bool SensorModel::simulateCellVision(int i, int j, GeoData* gdin, GeoData* gdout)
{ //{{{
	if (gdin == NULL || gdout == NULL)
	{
		return false;
	}
	double distFactor = 0.0;
	double cellScale  = gdin->getScale();
	int imax,imin;
	int jmax,jmin;
	int iscan,jscan;

	if (cellScale <= 0.0)
	{
		return false;
	}

	//--- Setting up the vision domain around (i,j)
	int radius = (int)(ceil(sensorRange/cellScale));	
	maxChanfreinDist = chanfreinDistance(0,0, radius,0);
	// Avoiding bad border effect 
	imin = i - radius;
	if (imin < 0 )
	{ imin = 0; }

	imax = i + radius;
	if (imax >= gdin->getxsize() )
	{ imax = gdin->getxsize() - 1; }

	// Avoiding bad border effect 
	jmin = j - radius;
	if (jmin < 0 )
	{ jmin = 0; }

	jmax = j + radius;
	if (jmax >= gdin->getysize() )
	{ jmax = gdin->getysize() - 1; }

	int katt = 0;
	int kgeo = GEODATA_TFLAT_BAND;
	double geovalue = 0.0;
	//--- Scanning pixels one by one and evaluating transformation
	Attributes tmpatt;
	Attributes tmpprior;
	for (iscan = imin; iscan <= imax; iscan++)
	{
		for (jscan = jmin; jscan <= jmax; jscan++)
		{
			distFactor = getDistanceFactor(i, j, iscan, jscan, gdin); 

			if (distFactor == 0.0)
			{ continue; }
			//--- Getting real world attributes at iscan,jscan
			kgeo = GEODATA_TFLAT_BAND;
			for (katt = ATT_BEGIN; katt <= T_LIMIT; katt++)
			{
				geovalue = gdin->getValueFromBand((GEODATA_BAND)kgeo, iscan, jscan);
				if (geovalue < 0.0)
				{
					geovalue = 0.0;
				}
				tmpatt.set((attributeKey)katt, geovalue);
				kgeo++;
			}   
			// Distributing certainties assuring normalization
			tmpatt.distributeCertainties();

			//--- Use estimated distance factor
			if (distFactor == -1.0)
			{
				// Close Range -  Simulate perfect vision
				gdout->setData(iscan, jscan, tmpatt, false);
			} else {
				//--- Getting prior knowledge
				kgeo = GEODATA_TFLAT_BAND;
				geovalue = 0.0;
				for (katt = ATT_BEGIN; katt <= T_LIMIT; katt++)
				{
					geovalue = gdout->getValueFromBand((GEODATA_BAND)kgeo, iscan, jscan);
					if (geovalue < 0.0)
					{
						geovalue = 0.0;
					}
					tmpprior.set((attributeKey)katt, geovalue);
					kgeo++;
				}
				// Distributing certainties assuring normalization
				tmpprior.distributeCertainties();

				//--- Apply model filter
				simulateAttributes(tmpatt, tmpprior, distFactor);
				//--- Writing tmpprior into the output geodata
				gdout->setData(iscan, jscan, tmpprior, false);
			}
		}
	}

	// Keeping internal geodata consistency as memToBuffer was disabled by
	// setData(,,false) to keep it fast.
	gdout->memToBuffer();

	return true;
} //}}}

bool SensorModel::simulateAttributes(const Attributes &attreal, Attributes& attprior, double distFactor)
{ //{{{
	if( !checkModelAppliance(attreal) )
	{ return false; }

	int k = 0;
	double tmp = 0.0;
	double tmpreal = 0.0;
	double randCoef = 0.0;
	double minVal = 0.001;
	Attributes attsimu;

	for (k = ATT_BEGIN; k < ATT_MAX; k++)
	{
		// Checking to which class this reality attribute belongs
		if ( attreal.certainty[(attributeKey)k] > 0.7 )
		{
			// attreal.certainty[(attributeKey)k] should be 1 if we're here.
			switch(randomModel)
			{
				case 1 :
					randCoef = 0.0;
				break;
				default: randCoef = rand()/(double)RAND_MAX;
			}

			//---- Applying a priori change according to the distance factor
			// Distance factor appliance domain
			tmp = (double)T_OBSTACLE - (double)k;
			if (tmp < (double)k)
			{
				tmp = (double)k;
			}
			// Spreading distribution a bit more than classes frontiers
			// When far from point of view the distribution is flatten.
			// 1.050 avoids zeros and big ones :) annoying in bayes filter.
			tmp = (1.050 + (1.0 - distFactor)) * tmp;

			//----- Calculating simulated distribution for perception input
			for (int m = ATT_BEGIN; m < ATT_MAX; m++)
			{
				// Repartition unnormal equation: will have to be normalized
				tmpreal = 0.0;
				if (tmp > 0.0)
				{
					tmpreal = fabs((double)k-(double)m)/tmp;
					tmpreal = 1.0 - (tmpreal*tmpreal);
					tmpreal = (tmpreal*tmpreal);
					tmpreal = (tmpreal*tmpreal);
				}
				attsimu.set((attributeKey)m, tmpreal);
			}
			attsimu.normalizeCertainties(ATT_BEGIN, ATT_END);

			//---- Applying sensor model and checking if it is mistaking
			// Random chance to be mistaking or not
			// Comparing the random coefficient with the P(Ci|Ci)
			if (randCoef <= sensormat[(k*sensormatSize + k)] && sensormat[(k*sensormatSize + k)] > 0.0 )
			{
				// Not mistaking - centered distribution around k
				// Building next prior ( == current aposteriori)
				//----- Quality damping
				for (int m = ATT_BEGIN; m < ATT_MAX; m++)
				{
					tmp = (attsimu.certainty[(attributeKey)m] * attprior.certainty[(attributeKey)m]);
					if (tmp <= minVal)
					{ tmp = minVal; }
					attprior.set((attributeKey)m, tmp);
				}
				attprior.normalizeCertainties(ATT_BEGIN, ATT_END);

				//------ Applying model
				//tmp = (sensormat[(k*sensormatSize + k)] - randCoef)/sensormat[(k*sensormatSize + k)];
				for (int m = ATT_BEGIN; m < ATT_MAX; m++)
				{
					tmp = sensormat[(k*sensormatSize + m)] * attprior.certainty[(attributeKey)m];
					if (tmp <= minVal)
					{ tmp = minVal; }
					attprior.set((attributeKey)m, tmp);
				}
				attprior.normalizeCertainties(ATT_BEGIN, ATT_END);

			} else {
				// Mistaking
				//----- Quality damping
				for (int m = ATT_BEGIN; m < ATT_MAX; m++)
				{
					tmp = ((1.0 - attsimu.certainty[(attributeKey)m]) * attprior.certainty[(attributeKey)m]);
					if (tmp <= minVal)
					{ tmp = minVal; }
					attprior.set((attributeKey)m, tmp);
				}
				attprior.normalizeCertainties(ATT_BEGIN, ATT_END);

				//------ Applying model
				// Random choice of the seen class
				tmp = (sensormat[(k*sensormatSize + k)] + ((1.0 - sensormat[(k*sensormatSize + k)])/2.0));
				if (  randCoef  <=  tmp  ) 
				{
					tmp = ((1.0 - sensormat[(k*sensormatSize + k)])/2.0);
					k = k + ((randCoef - sensormat[(k*sensormatSize + k)])/tmp) * (ATT_END - k);
				} else {
					tmp = ((1.0 - sensormat[(k*sensormatSize + k)])/2.0);
					k = k - ((randCoef - sensormat[(k*sensormatSize + k)] - tmp)/tmp) * (k - ATT_BEGIN);
				}

				// Caring about borders
				if (k > ATT_END)
				{ k = ATT_BEGIN; }
				if (k < ATT_BEGIN)
				{ k = ATT_END; }

				// Applying random mistake model
				for (int m = ATT_BEGIN; m < ATT_MAX; m++)
				{
					tmp = sensormat[(k*sensormatSize + m)] * attprior.certainty[(attributeKey)m];
					if (tmp <= minVal)
					{ tmp = minVal; }
					attprior.set((attributeKey)m, tmp);
				}
				attprior.normalizeCertainties(ATT_BEGIN, ATT_END);
			}

		} // end of if detected of which class is the concerned zone.
	}
	return true;
} //}}}

bool SensorModel::watchCell(int i, int j, GeoData* gdin, GeoData* gdout)
{ //{{{
	return true;
} //}}}

bool SensorModel::watchAttributes(const Attributes &attin, Attributes& attout)
{ //{{{
	if( !checkModelAppliance(attin) )
	{ return false; }

	int i = 0;
	int k = 0;
	double tmp = 0.0;
	for (k = ATT_BEGIN; k < ATT_MAX; k++)
	{
		// Suming all "probabilistic perceptions" that do not mistakes on the
		// case where terrain is of type k and all perceptions that mistakes
		// giving case k whereas the terrain is not.
		tmp = 0.0;
		for ( i = k*sensormatSize; 
		         (i < (k*sensormatSize+sensormatSize)) 
			  && (i < (sensormatSize*sensormatSize));
			  i++)
		{
			tmp = tmp + sensormat[i] * attin.certainty[(attributeKey)k];
		}
		attout.set((attributeKey)k, tmp);
	}

	attout.normalizeCertainties(ATT_BEGIN, T_OBSTACLE);
	return true;
} //}}}

bool SensorModel::checkModelAppliance(const Attributes &attin)
{ //{{{
	if (sensormatSize == ATT_MAX)
	{
		return true;
	}
	return false;
} //}}}

double SensorModel::chanfreinDistance(int ao, int bo, int a, int b)
{ //{{{
    int astep=0, bstep=0;
    double dist=0.0;
    /*------- Taking absolute step value -------*/
    if(ao >= a)
    {    astep = ao - a; }
    else
    {    astep = a - ao; }

    if(bo >= b)
    {    bstep = bo - b; }
    else
    {    bstep = b - bo; }

    /*------- Distance calculus -------*/
    if(bstep >= astep )
    {
        astep = bstep * 2 + astep;
        dist = (double)astep;
        return dist;
    }

    /*---- if no return ----*/
    astep = astep * 2 + bstep;
    dist = (double)astep;
    return dist;
}/*}}}*/

double SensorModel::getDistanceFactor(int ox, int oy, int i, int j, GeoData* geoW)
{ //{{{
    double dist = 0.0;
    double maxdist = 0.0;

    /* get the chanfrein distance with pixel coordinates */
    dist = chanfreinDistance(ox, oy, i , j ); 

	/* To suggest that the perception is taken in short range */
	if (dist <= shortRangeChanfreinDist)
	{
		return -1.0;
	}

	/* To suggest out of range perception */
    if (dist >= maxChanfreinDist)
    {
		return 0.0;
    }

    /* Used in adjustment of maxChanfreinDist to avoid zero distance factor 
	 * which pollutes probabilistic filtering.
	 */
    maxdist = maxChanfreinDist * distanceDampingFactor;

    /* Model the distance damping */
    switch (distanceDampingModel)
    {
		/* Update doxygen documentation in corresponding .hpp
		 * to tell if new models are available.
		 */
		/*
		case 43 : 	if (isIntervisible( geoW, ox, oy, 0, i, j, 0))
					{
						dist = 1.0 - (dist/maxdist)*(dist/maxdist)*(dist/maxdist)*(dist/maxdist);
					} else {
						// perception is out of range 
						dist = 0.0;
					}
			break;
			*/
		case 42 : 	if (isIntervisible(geoW, ox, oy, i, j))
					{
						dist = 1.0 - (dist/maxdist);
					} else {
						// perception is out of range 
						dist = 0.0;
					}
			break;
		case 41 : 	if (isIntervisible(geoW, ox, oy, i, j))
					{
						dist = 1.0;
					} else {
						// perception is out of range 
						dist = 0.0;
					}
			break;
		case 3 : dist = 1.0 - (dist/maxdist)*(dist/maxdist)*(dist/maxdist)*(dist/maxdist);
			 break;
		case 2 : dist = 1.0 - (dist/maxdist);
			 break;
		case 1 : dist = 1.0; 
			 break;
		default:
			dist = (dist / maxdist);
			dist = 1.0 - dist*dist*dist*dist;
			dist = dist * dist;
    }


    return dist;
}//}}}

bool SensorModel::isLandDrivable(GeoData* geoWorld, int x1coord, int y1coord, int x2coord, int y2coord)
{//{{{
	if (geoWorld == NULL) { return false; }

	// Final return value that will be overwritten by the intervisibility calculation
	bool areturn = false;

	//--- Preparing for driving context
	double savedSimProbaLim  = geoWorld->getSimulationProbaVisionLimit();
	double savedVisMinRadius = getVisibilityMinRadius();
	// Not driving on obstacles zones with >= 70% certainty
	geoWorld->setSimulationProbaVisionLimit(0.3);
	// Check every move even very short paths
	setVisibilityMinRadius(0.0);

	//--- Check level of drivability//traversability 
	areturn = isIntervisible(geoWorld, x1coord, y1coord, x2coord, y2coord);

	//--- Restore context
	geoWorld->setSimulationProbaVisionLimit(savedSimProbaLim);
	setVisibilityMinRadius(savedVisMinRadius);

	// return intervisibility result
	return areturn;
}//}}}

bool SensorModel::isIntervisible(GeoData* geoWorld, int x1coord, int y1coord, int x2coord, int y2coord)
{//{{{
	if (geoWorld == NULL) { return false; }

	// Number of bad intersections calculated (see below)
	int nBadIntersects = 0;
	int nBadI1 = 0;
	int nBadI2 = 0;
	int nBadI3 = 0;
	// Number of diagonal intersection that have been met
	//int diagonalCases= 0;
	//int trickyDiagonal = 0;
	int pointsLooping = 0;

	/* Points are on the same plane are calculation is made in the same z
	 * equivalent to the 2D consideration.
	 */
	/* Dealing with simple case of right next door configurations */
	if (chanfreinDistance( x1coord,y1coord,  x2coord,y2coord ) <= visibilityMinRadius)
	{
		return true; 
	}

	/* We calculate the straight line equation (vision equation) between
	 * the two points.
	 * Then we count the number of bad intersections. These bad
	 * intersections happen between two cells with at least one
	 * representing an obstacle.
	 *
	 * As soon as nBadIntersects >= 2 intervisiblity is not possible.
	 */

	// Start point : A (cell center)
	double Ax = (double)x1coord + 0.5;
	double Ay = (double)y1coord + 0.5;

	// Center point of the viewed cell (cell center)
	double Cx = (double)x2coord + 0.5;
	double Cy = (double)y2coord + 0.5;

	// End points : three closest corners
	double Bx1 = 0.0;
	double By1 = 0.0;
	double Bx2 = 0.0;
	double By2 = 0.0;
	double Bx3 = 0.0;
	double By3 = 0.0;
	double Bx = 0.0;
	double By = 0.0;

	// Intersection points
	double Ix = 0.0;
	double Iy = 0.0;
	double Ixmin = 0.0;
	double Iymin = 0.0;
	double Ixmax = 0.0;
	double Iymax = 0.0;

	int verticalCase = 0;
	int pureDiagonal = 0;

	// Slope coefficient of the visibility straight line :
	double m = 0.0; 

	// Setting end point as Bx,By :
	// Looping On the three closest corners.
	if (Ax < Cx)
	{
		Bx1 = Cx - 0.5;
		Bx2 = Cx - 0.5;
		Bx3 = Cx + 0.5;
	} else {
		Bx1 = Cx + 0.5;
		Bx2 = Cx + 0.5;
		Bx3 = Cx - 0.5;
	}

	if( Ay < Cy )
	{
		By1 = Cy - 0.5; 
		By2 = Cy + 0.5; 
		By3 = Cy - 0.5;
	} else {
		By1 = Cy + 0.5; 
		By2 = Cy - 0.5; 
		if (Ay == Cy)
		{
			// No third point considered.
			Bx3 = -1.0;
			By3 = -1.0;
			nBadI3 = 10;
		} else {
			By3 = Cy + 0.5;
		}
	}

	pointsLooping = 0;
	while (pointsLooping < 3)
	{
		if (pointsLooping==0)
		{
			Bx = Bx1;
			By = By1;
		}

		if (pointsLooping==1)
		{
			Bx = Bx2;
			By = By2;
			if (nBadI3 != 0) // then (bx3,by3) does not exists
			{
				// Avoiding looping on next last and third point
				pointsLooping = 4;
			}
		}

		if (pointsLooping==2)
		{
			Bx = Bx3;
			By = By3;
		}

		//-------- Starting intervisiblity counting
		nBadIntersects = 0;

		// Straight line equation : y = m x + b simplified to y = m x
		if (Ax != Bx)
		{
			m = (Ay-By) / (Ax-Bx);
		} else {
			// else we are in the case of the vertical line
			verticalCase = 1;
		}

		// Checking if we are on a pure diagonal
		if ( (Ax-Bx)*(Ax-Bx) == (Ay-By)*(Ay-By) )
		{
			pureDiagonal = 1;
		}

		// Defining intersection domain on X axis
		if (Ax < Bx)
		{
			Ixmin = Ax + 0.5;
			Ixmax = Bx;
		} else {
			Ixmin = Bx;
			Ixmax = Ax - 0.5;
		}

		// Defining intersection domain on Y axis
		if (Ay < By)
		{
			Iymin = Ay + 0.5;
			Iymax = By;
		} else {
			Iymin = By;
			Iymax = Ay - 0.5;
		}

		// Calculating bad intersections on X lines of the grid (verticals)
		if (Ax != Bx)
		{ //{{{
			Ix = Ixmin;
			while (Ix <= Ixmax && nBadIntersects < 2)
			{
				// Calculating intersection point
				Iy = Ay + (m * (Ix - Ax));

				// Intersection point (Ix, Iy) is between horizontal cells
				//if ( pureDiagonal == 1 ) //(Ix == (double)((int)Ix)  && Iy == (double)((int)Iy)) )
				if ( (Ix == (double)((int)Ix)  && Iy == (double)((int)Iy)) )
				{
					// Diagonal intersection case
					if (m < 0.0)
					{
						if ( !(    geoWorld->isVisionTransparent((int)Ix, ((int)Iy)-1)
								&& geoWorld->isVisionTransparent(((int)Ix)-1, ((int)Iy))  )  )
						{
							nBadIntersects++;
						}

						if (!(geoWorld->isVisionTransparent( (int)Ix,     (int)Iy ))
							&& !(geoWorld->isVisionTransparent( ((int)Ix)-1, ((int)Iy)-1 )  ))
						{
							nBadIntersects+=2;
						}
					} else {
						if ( !(    geoWorld->isVisionTransparent((int)Ix, ((int)Iy))
								&& geoWorld->isVisionTransparent(((int)Ix)-1, ((int)Iy)-1)  )  )
						{
							nBadIntersects++;
						}

						if (!(geoWorld->isVisionTransparent( (int)Ix,     ((int)Iy)-1 ))
							&& !(geoWorld->isVisionTransparent( ((int)Ix)-1, (int)Iy )))
							/*
							 * we have m >  0.0 ; meaning we have an ascending vision ray hitting point "o"
							 *       Ix
							 *    ---------
							 *    | # | ? |
							 *Iy  ----o----
							 *    | ? | # |
							 *    ---------
							 *
							 * #: obstructed cell
							 * =: free cell
							 * ?: we don't care
							 */
						{
							nBadIntersects+=2;
						}
					}
				} else {
					// Not a diagonal intersection 
					if (!( geoWorld->isVisionTransparent( (int)Ix,   (int)Iy )
						&& geoWorld->isVisionTransparent( ((int)Ix)-1, (int)Iy )) )
					{
						nBadIntersects++;
					}

				}

				Ix = Ix + 1.0;
			}
		} //}}}

		// Calculating bad intersections on Y lines of the grid (horizontals)
		if ( /* Ay != By */ (m != 0.0 || verticalCase == 1) && nBadIntersects < 2)
		{ //{{{
			Iy = Iymin;
			while (Iy <= Iymax && nBadIntersects < 2)
			{
				// Calculating intersection point
				if (m!=0.0 && verticalCase==0)
				{
					Ix = Ax + ((Iy - Ay)/m);
				} else {
					// Vertical case
					Ix = Ax; 
				}

				// Intersection point (Ix, Iy) is between vertical cells
				//if ( pureDiagonal == 1 ) //(Ix == (double)((int)Ix)  && Iy == (double)((int)Iy)) )
				if ( (Ix == (double)((int)Ix)  && Iy == (double)((int)Iy)) )
				{
					// These cases are already fully dealt while checking crossing points on X axis.
					// These points are the same now.
				} else {
					if (!(    geoWorld->isVisionTransparent( (int)Ix, (int)Iy)
						   && geoWorld->isVisionTransparent( (int)Ix, ((int)Iy)-1) ) )
					{
						nBadIntersects++;
					}
				}

				Iy = Iy + 1.0;
			}// eow
		} //}}} oeif

		if (pointsLooping == 0)
		{
			nBadI1 = nBadIntersects;
		}

		if (pointsLooping == 1 || pointsLooping > 3)
		{
			nBadI2 = nBadIntersects;
		}

		if (pointsLooping == 2)
		{
			nBadI3 = nBadIntersects;
		}
		pointsLooping++;
	}// eow pointsLooping

	if (nBadI1 >=2 && nBadI2 >= 2 && nBadI3 >= 2)
	{
		// No visibility with such counter states.
		return false;
	}

	return true;
}//}}}

double SensorModel::getVisibilityMinRadius()
{ //{{{
	return visibilityMinRadius;
} //}}}

//--------------------------------------------------------------------//

bool SensorModel::setSensorMatrix(int i, int j, double val)
{ //{{{
	int indexN = i*sensormatSize + j;
	if (indexN >= 0 && indexN < sensormatSize*sensormatSize)
	{
		sensormat[i*sensormatSize + j] = val;
		return true;
	}
	return false;
} //}}}

void SensorModel::setSensorRange(double ssr)
{ //{{{
	sensorRange = ssr;
} //}}}

void SensorModel::setDistanceDampingModel(int mod_)
{ //{{{
	distanceDampingModel = mod_;
} //}}}

void SensorModel::setVisibilityMinRadius(double vmr_)
{ //{{{
	visibilityMinRadius = vmr_;
} //}}}

//--------------------------------------------------------------------//

}}
